"""
This module defines a handler that forwards unparsed atoms to the event handlers.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

from aminer.input.InputInterfaces import AtomHandlerInterface
from aminer import AminerConfig
from aminer.parsing.MatchContext import DebugMatchContext


class SimpleUnparsedAtomHandler(AtomHandlerInterface):
    """Handlers of this class will just forward the received unparsed atoms to the registered event handlers."""

    def __init__(self, anomaly_event_handlers):
        """
        Initialise the Unparsed atom handler.
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        """
        super().__init__(anomaly_event_handlers=anomaly_event_handlers)

    def receive_atom(self, log_atom):
        """Receive an unparsed atom to create events for each."""
        if log_atom.is_parsed():
            return False
        try:
            data = log_atom.raw_data.decode(AminerConfig.ENCODING)
        except UnicodeError:
            data = repr(log_atom.raw_data)
        self.send_event_to_handlers(data, log_atom)
        return True

    def send_event_to_handlers(self, data, log_atom):
        """Send the data to the event handlers."""
        event_data = {}
        for listener in self.anomaly_event_handlers:
            listener.receive_event('Input.UnparsedAtomHandler', 'Unparsed atom received', [data], event_data, log_atom, self)


class VerboseUnparsedAtomHandler(SimpleUnparsedAtomHandler):
    """Handlers of this class will forward received unparsed atoms to the registered event handlers applying the DebugMatchContext."""

    def __init__(self, anomaly_event_handlers, parsing_model):
        """
        Initialise the Unparsed atom handler.
        @param anomaly_event_handlers for handling events, e.g., print events to stdout.
        """
        super().__init__(anomaly_event_handlers)
        self.parsing_model = parsing_model

    def send_event_to_handlers(self, data, log_atom):
        """Send the data to the event handlers."""
        match_context = DebugMatchContext(log_atom.raw_data)
        self.parsing_model.get_match_element('', match_context)
        debug_info = match_context.get_debug_info()
        debug_lines = []
        for line in debug_info.split('\n'):
            debug_lines.append(line.strip())
        event_data = {'DebugLog': debug_lines}
        for listener in self.anomaly_event_handlers:
            listener.receive_event(
                'Input.VerboseUnparsedAtomHandler', 'Unparsed atom received', [debug_info + data], event_data, log_atom, self)



from time import sleep
import socket
import sys
import os
sys.path.append('./')
sys.path.append('../../')
# skipcq: FLK-E402
from aminer.util.SecureOSFunctions import secure_open_file, send_annotated_file_descriptor

sock_name = '/tmp/test4unixSocket.sock'  # skipcq: BAN-B108
fd = secure_open_file(b'/etc/aminer/conf-enabled/Readme.txt', os.O_RDONLY)
sleep(0.5)
client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
client.connect(sock_name)
send_annotated_file_descriptor(client, fd, b'readmeStream', b'You should read these README instructions for better understanding.')



import unittest
from aminer.parsing.ParserMatch import ParserMatch
from aminer.parsing.MatchElement import MatchElement
from unit.TestBase import TestBase


class ParserMatchTest(TestBase):
    """Unittests for the ParserMatch."""

    match_element = MatchElement("path", b"match_string", b"match_object", None)

    def test1get_match_element(self):
        """Test if get_match_element works properly."""
        match = ParserMatch(self.match_element)
        self.assertEqual(match.get_match_element(), self.match_element)

    def test2get_match_dictionary(self):
        """Test if MatchElements with and without children are evaluated properly and if multiple calls are handled properly."""
        a3 = MatchElement("a3", b"a3", b"a3", None)
        a2 = MatchElement("a2", b"a2", b"a2", [a3])
        a1 = MatchElement("a1", b"a1", b"a1", [a2])
        b3 = MatchElement("b3", b"b3", b"b3", None)
        b2 = MatchElement("b2", b"b2", b"b2", [b3])
        b1 = MatchElement("b1", b"b1", b"b1", [b2])
        root_element = MatchElement("root", b"root", b"root", [a1, b1])

        parser_match = ParserMatch(root_element)
        dictionary = parser_match.get_match_dictionary()

        self.assertEqual(dictionary["root"], root_element)
        self.assertEqual(dictionary["a1"], a1)
        self.assertEqual(dictionary["a2"], a2)
        self.assertEqual(dictionary["a3"], a3)
        self.assertEqual(dictionary["b1"], b1)
        self.assertEqual(dictionary["b2"], b2)
        self.assertEqual(dictionary["b3"], b3)

    def test3match_element_input_validation(self):
        """Check if element_id is validated."""
        self.assertRaises(TypeError, ParserMatch, "string")
        self.assertRaises(TypeError, ParserMatch, None)
        self.assertRaises(TypeError, ParserMatch, b"path")
        self.assertRaises(TypeError, ParserMatch, 123)
        self.assertRaises(TypeError, ParserMatch, 123.22)
        self.assertRaises(TypeError, ParserMatch, True)
        self.assertRaises(TypeError, ParserMatch, {"id": "path"})
        self.assertRaises(TypeError, ParserMatch, ["path"])
        self.assertRaises(TypeError, ParserMatch, [])
        self.assertRaises(TypeError, ParserMatch, ())
        self.assertRaises(TypeError, ParserMatch, set())


if __name__ == "__main__":
    unittest.main()



from aminer.analysis.EventTypeDetector import EventTypeDetector
from aminer.analysis.VariableTypeDetector import VariableTypeDetector
from aminer.analysis.VariableCorrelationDetector import VariableCorrelationDetector
from aminer.input.LogAtom import LogAtom
from aminer.parsing.ParserMatch import ParserMatch
from aminer.parsing.MatchElement import MatchElement
from unit.TestBase import TestBase

from time import time
import random
from copy import deepcopy


class VariableCorrelationDetectorTest(TestBase):
    """This class containts unittests for the VariableCorrelationDetector."""

    # Number of execution of the tested function
    iterations = 20
    # Size of the initial datasample
    dataset_size = 100
    # Significance level
    significance_niveau = 0.05

    def test1filter_variables_with_vtd(self):
        """This test case checks if the variables are filtered accurately using the VariableTypeDetector."""
        self.filter_variables(True)

    def test2filter_variables_without_vtd(self):
        """This test case checks if the variables are filtered accurately without using the VariableTypeDetector."""
        self.filter_variables(False)

    def filter_variables(self, use_vtd):
        """Run the filter variables code with or without the VariableTypeDetector."""
        t = time()
        stat_data = b'5.3.0-55-generic'
        log_atom = LogAtom(stat_data, ParserMatch(MatchElement(None, stat_data, stat_data, None)), t, self.__class__.__name__)
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        if use_vtd:
            vtd = VariableTypeDetector(self.aminer_config, [self.stream_printer_event_handler], etd, num_init=self.dataset_size,
                                       div_thres=0.1, test_gof_int=True, sim_thres=0.3, gof_alpha=self.significance_niveau)
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1)
        for _ in range(self.dataset_size):
            etd.receive_atom(log_atom)
            if use_vtd:
                vtd.receive_atom(log_atom)
        vcd.init_cor(0)
        # the vcd should not learn any correlations in static data.
        self.assertEqual(vcd.pos_var_val, [[]])

        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        if use_vtd:
            vtd = VariableTypeDetector(self.aminer_config, [self.stream_printer_event_handler], etd, num_init=self.dataset_size,
                                       div_thres=0.1, test_gof_int=False, sim_thres=0.5, gof_alpha=self.significance_niveau)
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1)
        for i in range(self.dataset_size):
            stat_data = bytes(str((i % 60) * 0.1), 'utf-8')
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement("/", stat_data, stat_data, None)), t, self.__class__.__name__)
            etd.receive_atom(log_atom)
            if use_vtd:
                vtd.receive_atom(log_atom)
        vcd.init_cor(0)
        # the vcd should not learn any correlations in others data.
        self.assertEqual(vcd.pos_var_val, [[]])

        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        if use_vtd:
            vtd = VariableTypeDetector(self.aminer_config, [self.stream_printer_event_handler], etd, num_init=self.dataset_size,
                                       div_thres=0.1, test_gof_int=True, sim_thres=0.3, gof_alpha=self.significance_niveau)
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1)
        values = []
        for i in range(self.dataset_size):
            stat_data = bytes(str((i % 10) * 0.1), 'utf-8')
            values.append(float(stat_data))
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement("/", stat_data, stat_data, None)), t, self.__class__.__name__)
            etd.receive_atom(log_atom)
            if use_vtd:
                vtd.receive_atom(log_atom)
        vcd.init_cor(0)
        values_set = list(set(values))
        # the vcd should learn any correlations in discrete data.
        self.assertEqual(vcd.pos_var_val, [[values_set]])

        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        if use_vtd:
            vtd = VariableTypeDetector(self.aminer_config, [self.stream_printer_event_handler], etd, num_init=self.dataset_size,
                                       div_thres=0.1, test_gof_int=True, sim_thres=0.3, gof_alpha=self.significance_niveau)
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1)
        values = []
        for i in range(self.dataset_size):
            stat_data = bytes(str((i % 11) * 0.1), 'utf-8')
            values.append(float(stat_data))
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement(None, stat_data, stat_data, None)), t, self.__class__.__name__)
            etd.receive_atom(log_atom)
            if use_vtd:
                vtd.receive_atom(log_atom)
        vcd.init_cor(0)
        # the vcd should not learn any correlations if the discrete data is not in the threshold.
        self.assertEqual(vcd.pos_var_val, [[]])

    def test3initialize_variables_with_matchDiscDistr_preselection_method(self):
        """This test case checks the functionality of the matchDiscDistr preselection method."""
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1)
        values_list1 = [1.0/10]*10
        values_list2 = [1.0/14]*14
        # an correlation should be detected even if the second list contains more values than the first.
        self.assertTrue(vcd.pick_cor_match_disc_distr(values_list1, values_list2))

        values_list2 = [1.0/7]*7
        # an correlation should be detected even if the second list contains less values than the first.
        self.assertTrue(vcd.pick_cor_match_disc_distr(values_list1, values_list2))

        values_list2 = [1.0/30]*30
        # an correlation should not be detected if the probability of occurrence difference is too high.
        self.assertFalse(vcd.pick_cor_match_disc_distr(values_list1, values_list2))

        values_list2 = [0.2] + [0.8/9]*9
        # an correlation should not be detected if the probability of occurrence difference is too high.
        self.assertFalse(vcd.pick_cor_match_disc_distr(values_list1, values_list2))

        # find correlations even when the lists are randomly shuffled.
        values_list1 = [0.3]*2 + [0.4/3]*3
        values_list2 = [1.0/5] * 5
        random.shuffle(values_list1)
        self.assertTrue(vcd.pick_cor_match_disc_distr(values_list1, values_list2))

    def test4initialize_variables_with_excludeDueDistr_preselection_method(self):
        """This test case checks the functionality of the excludeDueDistr preselection method."""
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1)

        # equal distribution - no exclusion expected
        values = [0.1]*10
        self.assertTrue(vcd.pick_cor_exclude_due_distr(values))

        # almost equal distribution - no exclusion expected
        values = [0.3] + [0.078]*9
        self.assertTrue(vcd.pick_cor_exclude_due_distr(values))

        # one value with high probability - exclusion expected
        values = [0.5] + [0.056]*9
        self.assertFalse(vcd.pick_cor_exclude_due_distr(values))

        # multiple values with high probability - no exclusion expected
        values = [0.3]*3 + [0.014]*7
        self.assertTrue(vcd.pick_cor_exclude_due_distr(values))

        # check boundaries
        values = [0.5]*2
        self.assertTrue(vcd.pick_cor_exclude_due_distr(values))

        values = [0.8, 0.2]
        self.assertFalse(vcd.pick_cor_exclude_due_distr(values))

        values = [0.33]*3
        self.assertTrue(vcd.pick_cor_exclude_due_distr(values))

        values = [0.7] + [0.15]*2
        self.assertFalse(vcd.pick_cor_exclude_due_distr(values))

        values = [0.25]*4
        self.assertTrue(vcd.pick_cor_exclude_due_distr(values))

        values = [0.58] + [0.14]*3
        self.assertFalse(vcd.pick_cor_exclude_due_distr(values))

    def test5initialize_variables_with_matchDiscVals_preselection_method(self):
        """
        This test case checks the functionality of the matchDiscVals preselection method.
        This test actually uses values instead of probabilities, but they are similar to the values used in test3.
        """
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1)
        values_set1 = [i*0.1 for i in range(10)]
        values_set2 = [i*0.2 for i in range(7)]
        # an correlation should be detected even if the second list contains less values than the first.
        self.assertTrue(vcd.pick_cor_match_disc_vals(values_set1, values_set2))

        values_set2 = [i*0.3 for i in range(7)]
        # an correlation should not be detected if too many values are different.
        self.assertFalse(vcd.pick_cor_match_disc_vals(values_set1, values_set2))

        values = []
        for i in range(58):
            stat_data = bytes(str(i * 0.1), 'utf-8')
            values.append(float(stat_data))
        values_set1 = values

        values = []
        for i in range(41):
            stat_data = bytes(str(i * 0.2), 'utf-8')
            values.append(float(stat_data))
        values_set2 = values
        # an correlation should be detected if not too many values are different.
        self.assertTrue(vcd.pick_cor_match_disc_vals(values_set1, values_set2))

        values = []
        for i in range(42):
            stat_data = bytes(str(i * 0.2), 'utf-8')
            values.append(float(stat_data))
        values_set2 = values
        # an correlation should not be detected if too many values are different.
        self.assertFalse(vcd.pick_cor_match_disc_vals(values_set1, values_set2))

    def test6initialize_variables_with_random_preselection_method(self):
        """
        This test case checks the functionality of the random preselection method. It tests all percentage_random_cors in [0.01..1.0[.
        For all paths the possible amount of combinations is 10. The expected number of correlations is rounded. For example with
        0.05 <= percentage_random_cors < 0.15 exactly one combination is expected. The combinations also must not be repeated reversed and
        combinations with itself are not allowed. The used discrete data is for every path the same.
        """
        t = time()
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1,
                                          used_presel_meth=['random'])
        values = []
        for i in range(self.dataset_size):
            stat_data = bytes(str((i % 10) * 0.1), 'utf-8')
            values.append(float(stat_data))
            children = [MatchElement(str(j), stat_data, stat_data, None) for j in range(5)]
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement('/', str(i).encode(), str(i).encode(), children)),
                               t, self.__class__.__name__)
            etd.receive_atom(log_atom)
        vcd.init_cor(0)
        # test random correlation picking by using vcd.percentage_random_cors [0.01..1.0[
        for i in range(1, 100):
            vcd.percentage_random_cors = i / 100
            # out of 10 possible combinations exactly x should occur.
            x = i // 10 + (i % 10 >= 5)
            correlations = vcd.pick_cor_random(0)
            self.assertEqual(len(correlations), x, "Error at i = %d" % i)
            for corr in correlations:
                # one path must not correlate with itself.
                self.assertNotEqual(corr[0], corr[1])
                # the same, reversed combination must not be in values.
                self.assertFalse([corr[1], corr[0]] in correlations)

        # test if a ValueError is raised when percentage_random_cors is out of range.
        self.assertRaises(ValueError, VariableCorrelationDetector, self.aminer_config, [self.stream_printer_event_handler], etd,
                          disc_div_thres=0.1, used_presel_meth=['random'], percentage_random_cors=1.2)
        self.assertRaises(ValueError, VariableCorrelationDetector, self.aminer_config, [self.stream_printer_event_handler], etd,
                          disc_div_thres=0.1, used_presel_meth=['random'], percentage_random_cors=1.0)
        self.assertRaises(ValueError, VariableCorrelationDetector, self.aminer_config, [self.stream_printer_event_handler], etd,
                          disc_div_thres=0.1, used_presel_meth=['random'], percentage_random_cors=0.0)
        self.assertRaises(ValueError, VariableCorrelationDetector, self.aminer_config, [self.stream_printer_event_handler], etd,
                          disc_div_thres=0.1, used_presel_meth=['random'], percentage_random_cors=-1.2)

    def test7initialize_variables_with_intersect_presel_meth(self):
        """
        This test case checks the functionality of the intersect_presel_meth flag with multiple preselection methods.
        These are 'excludeDueDistr' and 'matchDiscVals'. In the first case intersect_presel_meth=False and correlations can be detected
        successfully. In the second case intersect_presel_meth=True and no correlations are found because they are excluded in
        'excludeDueDistr'.
        """
        t = time()
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd_union = VariableCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1, used_presel_meth=[
                'excludeDueDistr', 'matchDiscVals'], intersect_presel_meth=False)
        vcd_intersection = VariableCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5, used_presel_meth=[
                'excludeDueDistr', 'matchDiscVals'], intersect_presel_meth=True)
        vcd_exclude = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1,
                                                  used_presel_meth=['excludeDueDistr'])
        vcd_match = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1,
                                                used_presel_meth=['matchDiscVals'])
        var1 = ['a']*50 + ['b']*50
        var2 = ['a']*90 + ['b']*10
        var3 = ['c']*20 + ['d']*50 + ['e']*30
        var4 = ['c']*50 + ['d']*50

        for i, val in enumerate(var1):
            children = [MatchElement('2', var2[i].encode(), var2[i].encode(), None),
                        MatchElement('3', var3[i].encode(), var3[i].encode(), None),
                        MatchElement('4', var4[i].encode(), var4[i].encode(), None)]
            log_atom = LogAtom(val.encode(), ParserMatch(MatchElement('/', val.encode(), val.encode(), children)), t,
                               self.__class__.__name__)
            etd.receive_atom(log_atom)
        vcd_union.init_cor(0)
        vcd_intersection.init_cor(0)
        vcd_exclude.init_cor(0)
        vcd_match.init_cor(0)
        values_set = [[list(set(var1))] + [list(set(var2))] + [list(set(var3))] + [list(set(var4))]]
        self.assertTrue(sorted(vcd_union.pos_var_val), sorted(values_set))
        # intersect_presel_meth=False -> correlations should be found.
        # the correlation has to be in at least one presel method. (OR-Statement)
        unique_list = deepcopy(vcd_exclude.pos_var_cor[0])
        for cor in vcd_match.pos_var_cor[0]:
            if cor not in unique_list:
                unique_list.append(cor)
        self.assertEqual(len(unique_list), len(vcd_union.pos_var_cor[0]))

        values_set = [[list(set(var1))] + [list(set(var2))] + [list(set(var3))] + [list(set(var4))]]
        self.assertTrue(sorted(vcd_intersection.pos_var_val), sorted(values_set))
        # intersect_presel_meth=True -> correlations should still be found.
        # the correlation has to be in both presel methods. (AND-Statement)
        unique_list = []
        for cor in vcd_exclude.pos_var_cor[0]:
            if cor in vcd_match.pos_var_cor[0] and cor not in unique_list:
                unique_list.append(cor)
        for cor in vcd_match.pos_var_cor[0]:
            if cor in vcd_exclude.pos_var_cor[0] and cor not in unique_list:
                unique_list.append(cor)
        self.assertEqual(len(unique_list), len(vcd_intersection.pos_var_cor[0]))

    def test8initialize_variables_with_no_preselection_method(self):
        """
        This test case checks the selection with no preselection method used.
        Also this test case checks the functionality of the Rel and WRel methods. For the data generation the main path '/' always contains
        (i % 10)*1 and child elements contain (i % 10)*1 for half of the time and (i % 10)*2 for the other half. The first half of
        the data contains 10 different values. These values are not combined with other values like in the second half of the data, which
        introduces 5 new values. Therefore 15 combinations exist (5+4+3+2+1=15). 10 correlations exist when '/' = i*1 -> child = i*1. In the
        second half 5 new correlations are added when '/' = i*1 -> child = i*2.
        """
        t = time()
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5,
                                          num_init=self.dataset_size)
        values1 = []
        # generate the first half of the data with child elements being (i % 10) * 0.1.
        for i in range(self.dataset_size // 2):
            stat_data = bytes(str((i % 10) * 1), 'utf-8')
            values1.append(float(stat_data))
            children = [MatchElement(str(0), stat_data, stat_data, None)]
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement('/', str((i % 10) * 1).encode(), str((i % 10) * 1).encode(), children)),
                               t, self.__class__.__name__)
            etd.receive_atom(log_atom)
        values2 = []
        # generate the second half of the data with child elements being (i % 10) * 2.
        for i in range(self.dataset_size // 2):
            stat_data = bytes(str((i % 10) * 2), 'utf-8')
            values2.append(float(stat_data))
            children = [MatchElement(str(0), stat_data, stat_data, None)]
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement('/', str((i % 10) * 1).encode(), str((i % 10) * 1).encode(), children)),
                               t, self.__class__.__name__)
            etd.receive_atom(log_atom)
        vcd.init_cor(0)
        values_set = list(set(values1 + values2))
        pos_var_val = deepcopy(vcd.pos_var_val)
        # all child elements should contain data from values1 and values2.
        index = pos_var_val[0].index(values_set)
        del pos_var_val[0][index]
        # no other element should contain the united set of values1 and values2.
        self.assertRaises(ValueError, pos_var_val[0].index, values_set)
        # only values1 should be found, because the main path contains only data generated with (i % 10) * 1.
        self.assertEqual(pos_var_val, [[list(set(values1))]])

        # test the functionality of the Rel and WRel methods
        # copy both lists to not modify the actual lists of the vcd.
        rel_list = deepcopy(vcd.rel_list)
        w_rel_list = deepcopy(vcd.w_rel_list)
        for rel in rel_list[0]:
            for r in rel:
                step = 2
                for i in range(len(r)):  # skipcq: PTC-W0060
                    key = (i % 20 >= 10)*10 + ((i % 10) * step)
                    # search for the key k in the relation r or convert key to float if applicable.
                    for k in r:
                        if key == 0.0:
                            break
                        if k != 0.0 and k % key == 0:
                            key = k
                            break
                    value = r[key]
                    # there is no difference between the first half and the second half of the data, when value = 0.
                    if key == 0.0:
                        self.assertEqual({key: 10}, value)
                    # as the Rel method can learn only one relation, the values should be 2, 4, 6 and 8 when the key is divisible
                    # by 2 and smaller than 10.
                    elif key % 2 == 0 and key < 10.0:
                        self.assertEqual({key: 4}, value)
                    # as the Rel method can learn only one relation, the values should be 2, 4, 6 and 8 when the key is divisible
                    # by 2 and greater or equal 10.
                    elif key % 2 == 0:
                        self.assertEqual({(key/2): 5}, value)
                    else:
                        raise ValueError('The %f: %f combination must not occur in Rel.' % (key, value))

        # relations should be found in both directions and the count should be equal.
        cnt_half = 0  # for example key = 18.0 -> inner key = 9.0
        cnt_double = 0  # for example key = 9.0 -> inner key = 18.0
        for w_rel in w_rel_list[0]:
            for r in w_rel:
                step = 1.0
                # search for the step size
                for k in r:
                    if k >= 10.0:
                        step = 2.0
                if step == 1.0:
                    cnt_half += 1
                else:
                    cnt_double += 1
                for i in range(len(r)):  # skipcq: PTC-W0060
                    key = (i % 20 >= 10)*10 + ((i % 10) * step)
                    value = r[key]
                    # there is no difference between the first half and the second half of the data, when value = 0.
                    if key == 0.0:
                        self.assertEqual({key: 10}, value)
                    # this if is only reached when step = 2.0.
                    elif key >= 10.0:
                        self.assertEqual({key/2: 5}, value)
                    elif step == 1.0:
                        self.assertEqual({key*2: 5, key: 5}, value)
                    elif step == 2.0:
                        self.assertEqual({key/2: 5, key: 5}, value)
                    else:
                        raise ValueError('The %f: %f combination must not occur in WRel.' % (key, value))
        self.assertEqual(cnt_half, 1)
        self.assertEqual(cnt_double, 1)

    def test9nonexistent_preselection_methods(self):
        """This test case checks if an error occurs, when using an nonexistent preselection method."""
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        self.assertRaises(ValueError, VariableCorrelationDetector, self.aminer_config, [self.stream_printer_event_handler], etd,
                          disc_div_thres=0.1, used_presel_meth=['nonexistentPreselMeth'])

    def test10nonexistent_correlation_methods(self):
        """This test case checks if an error occurs, when using an nonexistent correlation method or empty list."""
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        self.assertRaises(ValueError, VariableCorrelationDetector, self.aminer_config, [self.stream_printer_event_handler], etd,
                          disc_div_thres=0.1, used_cor_meth=['nonexistentCorDMeth'])

    def test11validate_correlation_rules_coverVals(self):
        """
        This test case checks the functionality of the coverVals validation method.
        The validate_cor_cover_vals_thres is tested in the interval [0.1..1.0]. The data consists mostly of (i % 10) * 1 and every 7th
        value the child elements use (i % 10) * 2 as the condition (i % 7 == 0 and i != 0) is met. Comparing the count of values with
        h*10, as h is used to get the steps with 10%. If the count is smaller than h*10, no value must be found.
        """
        t = time()
        # run test for every 10% of validate_cor_cover_vals_thres
        for h in range(1, 11, 1):
            etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
            vcd = VariableCorrelationDetector(
                self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5, used_validate_cor_meth=['coverVals'],
                validate_cor_cover_vals_thres=0.7, num_init=self.dataset_size)
            # set new validate_cor_cover_vals_thres
            vcd.validate_cor_cover_vals_thres = h*0.1
            # init and validate. This is needed as the ETD also needs to be initialized.
            for i in range(self.dataset_size):
                stat_data = str((i % 10)).encode()
                children = [MatchElement(str(0), stat_data, stat_data, None)]
                log_atom = LogAtom(stat_data, ParserMatch(MatchElement('/', stat_data, stat_data, children)), t, self.__class__.__name__)
                etd.receive_atom(log_atom)
            vcd.init_cor(0)
            vcd.rel_list = [[[{9.0: {9.0: 26}, 16.0: {16.0: 13}}, {9.0: {9.0: 26}, 16.0: {16.0: 13}}]]]
            vcd.w_rel_list = [[[{9.0: {9.0: 26}, 16.0: {16.0: 13, 8.0: 5}}, {9.0: {9.0: 26}, 16.0: {16.0: 13, 8.0: 5}}]]]
            vcd.pos_var_cor = [[[0, 1]]]
            old_rel_list = deepcopy(vcd.rel_list[0])
            old_w_rel_list = deepcopy(vcd.w_rel_list[0])
            vcd.validate_cor()
            self.assertEqual(len(old_rel_list), len(vcd.rel_list[0]))
            self.assertEqual(len(old_w_rel_list), len(vcd.w_rel_list[0]))
            for i, rel in enumerate(vcd.rel_list[0]):
                for r in old_rel_list[i]:
                    cnt = 0
                    for key in r:
                        for val in r[key]:
                            cnt += r[key][val]
                    # when the count is smaller than validate_cor_cover_vals_thres in percent, then there should not be any correlations.
                    # h must be multiplied by 10 as it represents 10% steps.
                    if cnt < h * 10:
                        for val in rel:
                            self.assertEqual({}, val)
                    else:
                        self.assertEqual(vcd.rel_list[0], old_rel_list)

            for i, rel in enumerate(vcd.w_rel_list[0]):
                for r in old_w_rel_list[i]:
                    cnt = 0
                    for key in r:
                        for val in r[key]:
                            cnt += r[key][val]
                    # when the count is smaller than validate_cor_cover_vals_thres in percent, then there should not be any correlations.
                    # h must be multiplied by 10 as it represents 10% steps.
                    if cnt < h * 10:
                        for val in rel:
                            self.assertEqual({}, val)
                    else:
                        self.assertEqual(vcd.w_rel_list[0], old_w_rel_list)

    def test12validate_correlation_rules_distinctDistr(self):
        """
        This test case checks the functionality of the distinctDistr validation method.
        The first collection of datasets is similar and therefore produces more correlations. The second collection of datasets is not so
        similar and the number of correlations is smaller. The expected correlations can not be compared directly, because the order of the
        correlations is not guaranteed with the distinctDistr validation method. To achieve the equality test, both correlation variables
        are compared to [[], [], []] after all existing correlations are removed.
        """
        t = time()
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1, used_validate_cor_meth=['distinctDistr'],
            validate_cor_distinct_thres=0.05, num_init=self.dataset_size)
        # init and validate
        similar_data1 = ['a']*50 + ['b']*20 + ['c']*25 + ['d']*5
        similar_data2 = ['a']*45 + ['b']*25 + ['c']*15 + ['d']*10 + ['e']*5
        similar_data3 = ['a']*55 + ['b']*15 + ['c']*20 + ['d']*10
        unsimilar_data1 = ['a']*50 + ['b']*20 + ['c']*25 + ['d']*5
        unsimilar_data2 = ['a']*10 + ['b']*15 + ['c']*15 + ['d']*10 + ['e']*50
        unsimilar_data3 = ['a']*25 + ['b']*15 + ['c']*50 + ['d']*10

        for i in range(self.dataset_size):
            children = [MatchElement(str(1), similar_data2[i].encode(), similar_data2[i].encode(), None),
                        MatchElement(str(2), similar_data3[i].encode(), similar_data3[i].encode(), None)]
            log_atom = LogAtom(similar_data1, ParserMatch(MatchElement(
                '/', similar_data1[i].encode(), similar_data1[i].encode(), children)), t, self.__class__.__name__)
            etd.receive_atom(log_atom)
        vcd.init_cor(0)
        old_w_rel_list = deepcopy(vcd.w_rel_list[0])
        vcd.validate_cor()
        self.assertEqual(len(old_w_rel_list), len(vcd.w_rel_list[0]))
        expected_similar_correlations = [[{
            'd': {'e': 5}, 'c': {'d': 10, 'c': 15}, 'b': {'b': 20}, 'a': {'b': 5, 'a': 45}}, {
            'e': {'d': 5}, 'd': {'c': 10}, 'c': {'c': 15}, 'b': {'b': 20, 'a': 5}, 'a': {'a': 45}}], [{
                'd': {'d': 5}, 'c': {'d': 5, 'c': 20}, 'b': {'b': 15, 'a': 5}, 'a': {'a': 50}}, {
                'd': {'d': 5, 'c': 5}, 'c': {'c': 20}, 'b': {'b': 15}, 'a': {'b': 5, 'a': 50}}], [{
                    'e': {'d': 5}, 'd': {'d': 5, 'c': 5}, 'c': {'c': 15}, 'b': {'b': 15, 'a': 10}, 'a': {'a': 45}},
            {'d': {'e': 5, 'd': 5}, 'c': {'d': 5, 'c': 15}, 'b': {'b': 15}, 'a': {'b': 10, 'a': 45}}]]
        for w_rel in vcd.w_rel_list[0]:
            for cor in w_rel:
                deleted = False
                for expected_similar_correlation in expected_similar_correlations:
                    if cor in expected_similar_correlation:
                        index = expected_similar_correlation.index(cor)
                        del expected_similar_correlation[index]
                        deleted = True
                        break
                # if the correlation was not deleted an error is raised and the test fails.
                if not deleted:
                    raise ValueError('Correlation %s could not be found in the WRel List.' % cor)
        self.assertEqual([[], [], []], expected_similar_correlations)

        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vtd = VariableTypeDetector(self.aminer_config, [self.stream_printer_event_handler], etd, num_init=self.dataset_size, div_thres=0.1,
                                   test_gof_int=True, sim_thres=0.1, gof_alpha=self.significance_niveau)
        vcd = VariableCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.1, used_validate_cor_meth=['distinctDistr'],
            validate_cor_distinct_thres=0.05, num_init=self.dataset_size)
        for i in range(self.dataset_size):
            children = [MatchElement(str(1), unsimilar_data2[i].encode(), unsimilar_data2[i].encode(), None),
                        MatchElement(str(2), unsimilar_data3[i].encode(), unsimilar_data3[i].encode(), None)]
            log_atom = LogAtom(unsimilar_data1[i], ParserMatch(MatchElement(
                '/', unsimilar_data1[i].encode(), unsimilar_data1[i].encode(), children)), t, self.__class__.__name__)
            etd.receive_atom(log_atom)
            vtd.receive_atom(log_atom)
        vcd.init_cor(0)
        old_w_rel_list = deepcopy(vcd.w_rel_list[0])
        vcd.validate_cor()
        self.assertEqual(len(old_w_rel_list), len(vcd.w_rel_list[0]))
        expected_unsimilar_correlations = [[
            {}, {'a': {'a': 10}, 'b': {'a': 15}, 'c': {'a': 15}, 'd': {'a': 10}, 'e': {'b': 20, 'c': 25, 'd': 5}}], [
            {}, {'a': {'a': 25}, 'b': {'a': 15}, 'd': {'c': 5, 'd': 5}}], [
            {'a': {'a': 10}, 'b': {'a': 15}, 'c': {'b': 15}, 'd': {'c': 10}, 'e': {'c': 40, 'd': 10}}, {
                'a': {'a': 10, 'b': 15}, 'b': {'c': 15}, 'c': {'d': 10, 'e': 40}, 'd': {'e': 10}}]]
        for w_rel in vcd.w_rel_list[0]:
            for cor in w_rel:
                deleted = False
                for expected_unsimilar_correlation in expected_unsimilar_correlations:
                    if cor in expected_unsimilar_correlation:
                        index = expected_unsimilar_correlation.index(cor)
                        del expected_unsimilar_correlation[index]
                        deleted = True
                        break
                # if the correlation was not deleted an error is raised and the test fails.
                if not deleted:
                    raise ValueError('Correlation %s could not be found in the WRel List.%s' % (cor, vcd.w_rel_list[0]))
        self.assertEqual([[], [], []], expected_unsimilar_correlations)

    def test13validate_correlation_rules_distinctDistr_without_WRel(self):
        """This test case checks if an error occurs, when using the distinctDistr validation method without the WRel correlation method."""
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        self.assertRaises(ValueError, VariableCorrelationDetector, self.aminer_config, [self.stream_printer_event_handler], etd,
                          disc_div_thres=0.1, used_cor_meth=['Rel'], used_validate_cor_meth=['distinctDistr'])

    def test14nonexistent_validation_method(self):
        """This test case checks if an error occurs, when using an nonexistent validation method or empty list."""
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        self.assertRaises(ValueError, VariableCorrelationDetector, self.aminer_config, [self.stream_printer_event_handler], etd,
                          disc_div_thres=0.1, used_validate_cor_meth=['nonexistentValidateCorDMeth'])

    def test15update_and_test_correlation_rules_with_rel_correlation_method(self):
        """
        This test case checks the functionality of the Rel correlation method in the update, correlation generation and test phases.
        The correlations are initialized with 10 values for each correlation and keys calculated with (i % 10) * 0.1. In the update phase
        keys are calculated with (i % 10) * 0.2. Due to that the existing value's count must stay the same in cases where new values are not
        created and new values must be created from 1.0 to 1.8. Values are increased by or created with a count of 10.
        """
        description = 'test15VCD1'
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5,
                                          used_cor_meth=['Rel'], num_init=self.dataset_size)
        self.analysis_context.register_component(vcd, description)
        self.update_or_test_with_rel_correlation_method(etd, vcd, update_rules=True, generate_rules=True)
        for rel in vcd.rel_list[0]:
            for r in rel:
                for i in r:
                    key = i
                    value = r[key]
                    # existing values which are divisible by 2 and smaller than 10.0 should be updated.
                    if key % 2 == 0 and key < 10.0:
                        self.assertEqual({key: 20}, value)
                    # new values which are divisible by 2 and greater than 10.0 should be created.
                    # other values should stay the same as before.
                    else:
                        self.assertEqual({key: 10}, value)

        description = 'test15VCD2'
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5,
                                          used_cor_meth=['Rel'], num_init=self.dataset_size)
        self.analysis_context.register_component(vcd, description)
        self.update_or_test_with_rel_correlation_method(etd, vcd, update_rules=True, generate_rules=False)
        for rel in vcd.rel_list[0]:
            for r in rel:
                for i in r:
                    key = i
                    value = r[key]
                    # no new values should be created.
                    self.assertFalse(key % 2 == 0 and key >= 10.0)
                    # existing values which are divisible by 2 and smaller than 10.0 should be updated.
                    if key % 2 == 0 and key < 10.0:
                        self.assertEqual({key: 20}, value)
                    # other values should stay the same as before.
                    else:
                        self.assertEqual({key: 10}, value)

        description = 'test15VCD3'
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5,
                                          used_cor_meth=['Rel'], num_init=self.dataset_size)
        self.analysis_context.register_component(vcd, description)
        old_rel_list = self.update_or_test_with_rel_correlation_method(etd, vcd, update_rules=False, generate_rules=False)
        # no values in the rel_list should be changed.
        self.assertEqual(vcd.rel_list[0], old_rel_list)

        description = 'test15VCD4'
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5,
                                          used_cor_meth=['Rel'], num_init=self.dataset_size)
        self.analysis_context.register_component(vcd, description)
        offset = 200
        self.update_or_test_with_rel_correlation_method(etd, vcd, update_rules=True, generate_rules=False, offset=offset)
        # old correlations from child elements with the value being divisible by 2 should be deleted. The first ten correlations from the
        # initialization phase were not touched and should remain the same. The other correlation however should delete every value which
        # is divisible by 2.
        rel_list = deepcopy(vcd.rel_list[0])
        # delete correlations from the init phase.
        for rel in vcd.rel_list[0]:
            if rel[0] == rel[1]:
                index = rel_list.index(rel)
                del rel_list[index]
        self.assertEqual(1, len(rel_list))
        for rel in rel_list:
            # the order of the correlations is not guaranteed.
            if len(rel[0]) > len(rel[1]):
                rel0 = rel[0]
                rel1 = rel[1]
            else:
                rel0 = rel[1]
                rel1 = rel[0]
            for i in rel0:
                key = i
                value = rel0[key]
                self.assertEqual({key: 10}, value)
            for i in rel1:
                key = i
                value = rel1[key]
                # no values divisible by 2 should exist.
                self.assertFalse(key % 2 == 0)
                self.assertEqual({key: 10}, value)

    def update_or_test_with_rel_correlation_method(self, etd, vcd, update_rules, generate_rules, offset=0):
        """Run the update or test of rel correlations."""
        t = time()
        values = []
        # generate the initialization data with child elements being (i % 10) * 1.
        for i in range(self.dataset_size):
            stat_data = bytes(str((i % 10) * 1), 'utf-8')
            values.append(float(stat_data))
            children = [MatchElement(str(0), stat_data, stat_data, None)]
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement('/', stat_data, stat_data, children)), t,
                               self.__class__.__name__)
            etd.receive_atom(log_atom)
        vcd.init_cor(0)
        # test if the initialization contains only correlations with 10 values.
        for rel in vcd.rel_list[0]:
            for r in rel:
                for i in r:
                    key = i
                    value = r[key]
                    self.assertEqual({key: 10}, value)
        old_rel_list = deepcopy(vcd.rel_list[0])
        values = []
        # generate the update data with child elements being (i % 10) * 2.
        for i in range(self.dataset_size):
            stat_data = bytes(str((i % 10) * 2 + offset), 'utf-8')
            values.append(float(stat_data))
            children = [MatchElement(str(0), stat_data, stat_data, None)]
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement('/', str((i % 10) * 2).encode(), str((i % 10) * 2).encode(), children)),
                               t, self.__class__.__name__)
            etd.receive_atom(log_atom)
        vcd.log_atom = log_atom
        vcd.update_rules[0] = update_rules
        vcd.generate_rules[0] = generate_rules
        vcd.update_or_test_cor(0)
        return old_rel_list

    def test16update_and_test_correlation_rules_with_w_rel_correlation_method(self):
        """
        This test case checks the functionality of the WRel correlation method in the update, correlation generation and test phases.
        The correlations are initialized with 70% of the values having (i % 10) * 0.1 and 30% of the values having (i % 10) * 0.2. In the
        update phase the ratio is changed from 70:30 to 80:20. Thus the expected ratio is 75:25, when update_rules=True wihout offset.
        """
        # This part tests if rules are updated when update_rules=True and generate_rules=True, however no new rules are generated as the
        # same data is passed on in the update process.
        description = 'test16VCD1'
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5, used_cor_meth=['WRel'],
            num_init=self.dataset_size, num_update=self.dataset_size, max_dist_rule_distr=0.5)
        self.analysis_context.register_component(vcd, description)
        self.update_or_test_with_w_rel_correlation_method(etd, vcd, update_rules=True, generate_rules=True)
        self.assertEqual(1, len(vcd.w_rel_list[0]))
        for rel in vcd.w_rel_list[0]:
            for r in rel:
                for i in r:
                    key = i
                    value = r[key]
                    if key == 0:
                        self.assertEqual({key: 20}, value)
                    elif key >= 10.0:
                        self.assertEqual({key/2: 5}, value)
                    elif key % 2 == 0:
                        self.assertTrue(value in ({key/2: 5, key: 15}, {key*2: 5, key: 15}))
                    else:
                        self.assertTrue(value in ({key: 15}, {key*2: 5, key: 15}))

        # This part tests if rules are updated when update_rules=True and generate_rules=False. Therefore the assumptions of correlations is
        # the same as above, because there were no new correlations generated due to the same data being used.
        description = 'test16VCD2'
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5, used_cor_meth=['WRel'],
            num_init=self.dataset_size, num_update=self.dataset_size, max_dist_rule_distr=0.5)
        self.analysis_context.register_component(vcd, description)
        self.update_or_test_with_w_rel_correlation_method(etd, vcd, update_rules=True, generate_rules=False)
        self.assertEqual(1, len(vcd.w_rel_list[0]))
        for rel in vcd.w_rel_list[0]:
            for r in rel:
                for i in r:
                    key = i
                    value = r[key]
                    if key == 0:
                        self.assertEqual({key: 20}, value)
                    elif key >= 10.0:
                        self.assertEqual({key/2: 5}, value)
                    elif key % 2 == 0:
                        self.assertTrue(value in ({key/2: 5, key: 15}, {key*2: 5, key: 15}))
                    else:
                        self.assertTrue(value in ({key: 15}, {key*2: 5, key: 15}))

        # This part tests if rules are updated when update_rules=False and generate_rules=False. No correlation should be changed.
        description = 'test16VCD3'
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5, used_cor_meth=['WRel'],
            num_init=self.dataset_size, num_update=self.dataset_size, max_dist_rule_distr=0.5)
        self.analysis_context.register_component(vcd, description)
        old_w_rel_list = self.update_or_test_with_w_rel_correlation_method(etd, vcd, update_rules=False, generate_rules=False)
        # no values in the rel_list should be changed.
        self.assertEqual(vcd.w_rel_list[0], old_w_rel_list)

        # This part tests if rules are updated when update_rules=True and generate_rules=False but with an offset of 200. Therefore the
        # assumptions of correlations for the first part should stay the same and no new correlations should be learned, because an offset
        # is added to all data.
        description = 'test16VCD4'
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5, used_cor_meth=['WRel'],
            num_init=self.dataset_size, num_update=self.dataset_size, max_dist_rule_distr=0.5)
        self.analysis_context.register_component(vcd, description)
        offset = 200
        self.update_or_test_with_w_rel_correlation_method(etd, vcd, update_rules=True, generate_rules=False, offset=offset)
        self.assertEqual(1, len(vcd.w_rel_list[0]))
        for rel in vcd.w_rel_list[0]:
            for r in rel:
                for i in r:
                    key = i
                    value = r[key]
                    if key == 0:
                        self.assertTrue(value in ({key: 10}, {key: 10, float(offset): 2}))
                    elif key >= 10.0:
                        self.assertEqual({key / 2: 3}, value)
                    elif key % 2 == 0:
                        self.assertTrue(value in ({key/2: 3, key: 7}, {key*2: 3, key: 7, key*2+offset: 2}))
                    else:
                        self.assertTrue(value in ({key: 7}, {key*2: 3, key: 7, key*2+offset: 2}))

        # This part tests if rules are updated when update_rules=True and generate_rules=True but with an offset of 200. Therefore the
        # assumptions of correlations for the first part should stay the same and new correlations should be learned, because an offset
        # is added to all data.
        description = 'test16VCD5'
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(
            self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5,
            used_cor_meth=['WRel'], num_init=self.dataset_size, num_update=self.dataset_size, max_dist_rule_distr=0.5)
        self.analysis_context.register_component(vcd, description)
        offset = 200
        self.update_or_test_with_w_rel_correlation_method(etd, vcd, update_rules=True, generate_rules=True, offset=offset)
        self.assertEqual(1, len(vcd.w_rel_list[0]))
        for rel in vcd.w_rel_list[0]:
            for r in rel:
                for i in r:
                    key = i
                    value = r[key]
                    if key == 0:
                        self.assertTrue(value in ({key: 10}, {key: 0, float(offset): 2}))
                    elif key >= 10.0:
                        self.assertTrue(value in ({(key-offset)/2: 2}, {key/2: 3}, {(key-offset)/2: 2, key: 8}, {key: 8}))
                    elif key % 2 == 0:
                        self.assertTrue(value in ({key/2: 3, key: 7}, {key*2: 0, key: 0, key*2+offset: 2}))
                    else:
                        self.assertTrue(value in ({key: 7}, {key*2: 0, key: 0, key*2+offset: 2}))

    def update_or_test_with_w_rel_correlation_method(self, etd, vcd, update_rules, generate_rules, offset=0):
        """
        Run the update or test of w_rel correlations. This method initializes the vcd with a distribution of 70% 0.1 and 30% 0.2.
        In the update phase the distribution is 80% 1 and 20% 2.
        """
        t = time()
        values = []
        # generate the initialization data with child elements being (i % 10) * 1.
        for i in range(70):
            stat_data = bytes(str((i % 10) * 1), 'utf-8')
            values.append(float(stat_data))
            children = [MatchElement(str(0), stat_data, stat_data, None)]
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement('/', stat_data, stat_data, children)), t,
                               self.__class__.__name__)
            etd.receive_atom(log_atom)
        for i in range(30):
            stat_data = bytes(str((i % 10) * 2), 'utf-8')
            values.append(float(stat_data))
            children = [MatchElement(str(0), stat_data, stat_data, None)]
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement('/', str((i % 10) * 1).encode(), str((i % 10) * 1).encode(), children)),
                               t, self.__class__.__name__)
            etd.receive_atom(log_atom)
        vcd.init_cor(0)
        old_w_rel_list = deepcopy(vcd.w_rel_list[0])
        self.assertEqual(1, len(vcd.w_rel_list[0]))
        for rel in vcd.w_rel_list[0]:
            for r in rel:
                for i in r:
                    key = i
                    value = r[key]
                    if key == 0:
                        self.assertEqual({key: 10}, value)
                    elif key >= 10.0:
                        self.assertEqual({key/2: 3}, value)
                    elif key % 2 == 0:
                        self.assertTrue(value in ({key/2: 3, key: 7}, {key*2: 3, key: 7}))
                    else:
                        self.assertTrue(value in ({key: 7}, {key*2: 3, key: 7}))

        values = []
        for i in range(80):
            stat_data = bytes(str((i % 10) * 1 + offset), 'utf-8')
            values.append(float(stat_data))
            children = [MatchElement(str(0), stat_data, stat_data, None)]
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement('/', stat_data, stat_data, children)), t,
                               self.__class__.__name__)
            etd.receive_atom(log_atom)
        for i in range(20):
            stat_data = bytes(str((i % 10) * 2 + offset), 'utf-8')
            values.append(float(stat_data))
            children = [MatchElement(str(0), stat_data, stat_data, None)]
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement('/', str((i % 10) * 1).encode(), str((i % 10) * 1).encode(), children)),
                               t, self.__class__.__name__)
            etd.receive_atom(log_atom)
        vcd.log_atom = log_atom
        vcd.update_rules[0] = update_rules
        vcd.generate_rules[0] = generate_rules
        vcd.update_or_test_cor(0)
        return old_w_rel_list

    def test17init_and_update_timings(self):
        """This test checks if the init and update intervals are calculated correctly."""
        description = 'test17VCD1'
        t = time()
        etd = EventTypeDetector(self.aminer_config, [self.stream_printer_event_handler])
        vcd = VariableCorrelationDetector(self.aminer_config, [self.stream_printer_event_handler], etd, disc_div_thres=0.5,
                                          num_init=self.dataset_size, num_update=self.dataset_size)
        self.analysis_context.register_component(vcd, description)
        values = []
        for i in range(self.dataset_size):
            stat_data = bytes(str((i % 10) * 0.1), 'utf-8')
            values.append(float(stat_data))
            children = [MatchElement(str(0), stat_data, stat_data, None)]
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement(
                '/', str((i % 10) * 0.1).encode(), str((i % 10) * 0.1).encode(), children)), t, self.__class__.__name__)
            etd.receive_atom(log_atom)
            vcd.receive_atom(log_atom)
            if i < self.dataset_size - 1:
                self.assertEqual(vcd.pos_var_cor, [])
                self.assertEqual(vcd.pos_var_val, [])
                self.assertEqual(vcd.w_rel_list, [])
                self.assertEqual(vcd.rel_list, [])
        # just check if some values were learned and save them to compare.
        self.assertNotEqual(vcd.pos_var_cor, [])
        self.assertNotEqual(vcd.pos_var_val, [])
        self.assertNotEqual(vcd.w_rel_list, [])
        self.assertNotEqual(vcd.rel_list, [])
        old_pos_var_cor = deepcopy(vcd.pos_var_cor)
        old_pos_var_val = deepcopy(vcd.pos_var_val)
        old_w_rel_list = deepcopy(vcd.w_rel_list)
        old_rel_list = deepcopy(vcd.rel_list)

        values = []
        for i in range(self.dataset_size):
            stat_data = bytes(str((i % 10) * 1), 'utf-8')
            values.append(float(stat_data))
            children = [MatchElement(str(0), stat_data, stat_data, None)]
            log_atom = LogAtom(stat_data, ParserMatch(MatchElement('/', str((i % 10) * 1).encode(), str((i % 10) * 1).encode(), children)),
                               t, self.__class__.__name__)
            etd.receive_atom(log_atom)
            vcd.receive_atom(log_atom)
            if i < self.dataset_size - 1:
                self.assertEqual(vcd.pos_var_cor, old_pos_var_cor)
                self.assertEqual(vcd.pos_var_val, old_pos_var_val)
                self.assertEqual(vcd.w_rel_list, old_w_rel_list)
                self.assertEqual(vcd.rel_list, old_rel_list)
        # no new values are expected as num_steps_create_new_rules is -1 by default.
        self.assertEqual(vcd.pos_var_cor, old_pos_var_cor)
        self.assertEqual(vcd.pos_var_val, old_pos_var_val)
        self.assertNotEqual(vcd.w_rel_list, old_w_rel_list)
        self.assertNotEqual(vcd.rel_list, old_rel_list)



{
    'Parser': {
        'required': True,
        'type': 'list',
        'has_start': True,
        'schema': {
            'type': 'dict',
            'schema': {
                'id': {'type': 'string', 'required': True},
                'start': {'type': 'boolean'},
                'type': {'type': 'parsermodel', 'coerce': 'toparsermodel', 'required': True},
                'name': {'type': 'string', 'required': True},
                'args': {'type': ['string', 'list'], 'schema': {'type': ['string', 'integer']}, 'nullable': True},
                'branch_model_dict': {'type': 'list', 'schema': {'type': 'dict', 'schema': {'id': {
                    'type': ['boolean', 'float', 'integer', 'string']}, 'model': {'type': 'string'}}}},
                'date_formats': {'type': 'list', 'schema': {'type': 'dict', 'schema': {'format': {'type': 'list', 'schema': {
                    'type': 'string', 'nullable': True}}}}},
                'value_sign_type': {'type': 'string', 'allowed': ['none', 'optional', 'mandatory'], 'default': 'none'},
                'value_pad_type': {'type': 'string', 'allowed': ['none', 'zero', 'blank'], 'default': 'none'},
                'exponent_type': {'type': 'string', 'allowed': ['none', 'optional', 'mandatory'], 'default': 'none'},
                'start_year': {'type': 'integer', 'nullable': True, 'default': None},
                'delimiter': {'type': 'string'},
                'escape': {'type': 'string', 'nullable': True, 'default': None},
                'consume_delimiter': {'type': 'boolean', 'default': False},
                'key_parser_dict': {'type': 'dict'},
                'optional_key_prefix': {'type': 'string', 'default': 'optional_key_'},
                'nullable_key_prefix': {'type': 'string', 'default': '+'},
                'strict': {'type': 'boolean', 'default': False},
                'ignore_null': {'type': 'boolean', 'default': True},
                'date_format': {'type': 'string', 'minlength': 2},
                'text_locale': {'type': 'string', 'nullable': True, 'default': None},
                'max_time_jump_seconds': {'type': 'integer', 'default': 86400},
                'timestamp_scale': {'type': 'integer', 'default': 1},
                'allow_all_fields': {'type': 'boolean', 'default': False}
            }
        }
    },
}



"""
This module defines functions for secure file handling.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
"""

import os
import socket
import struct
import sys
import logging
from aminer.AminerConfig import DEBUG_LOG_NAME


base_dir_fd = None
tmp_base_dir_fd = None
log_dir_fd = None
base_dir_path = None
tmp_base_dir_path = None
log_dir_path = None


def secure_open_base_directory(directory_name=None, flags=0):
    """Open the base directory in a secure way."""
    global base_dir_fd  # skipcq: PYL-W0603
    global base_dir_path  # skipcq: PYL-W0603
    global tmp_base_dir_fd  # skipcq: PYL-W0603
    global tmp_base_dir_path  # skipcq: PYL-W0603
    if directory_name is not None and isinstance(directory_name, str):
        directory_name = directory_name.encode()
    if base_dir_path is None and (directory_name is None or not directory_name.startswith(b'/')):
        msg = 'Secure open on relative path not supported'
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)
    if base_dir_path is None and (flags & os.O_DIRECTORY) == 0:
        msg = 'Opening directory but O_DIRECTORY flag missing'
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)

    if base_dir_fd is None:
        base_dir_fd = os.open(directory_name, flags | os.O_NOFOLLOW | os.O_NOCTTY | os.O_DIRECTORY)
        base_dir_path = directory_name
        tmp_base_dir_path = directory_name
        tmp_base_dir_fd = os.open(tmp_base_dir_path, flags | os.O_NOFOLLOW | os.O_NOCTTY | os.O_DIRECTORY)
    return base_dir_fd


def close_base_directory():
    """Close the base directory at program shutdown."""
    global base_dir_fd  # skipcq: PYL-W0603
    global tmp_base_dir_fd  # skipcq: PYL-W0603
    global base_dir_path  # skipcq: PYL-W0603
    try:
        if base_dir_fd is not None:
            os.close(base_dir_fd)
            base_dir_fd = None
            base_dir_path = None
        if tmp_base_dir_fd is not None:
            os.close(tmp_base_dir_fd)
            tmp_base_dir_fd = None
    except OSError as e:
        msg = 'Could not close the base directory. Error: %s' % e
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)


def secure_open_log_directory(log_directory_name=None, flags=0):
    """Open the base log directory in a secure way."""
    global log_dir_fd  # skipcq: PYL-W0603
    global log_dir_path  # skipcq: PYL-W0603
    if log_directory_name is not None and isinstance(log_directory_name, str):
        log_directory_name = log_directory_name.encode()
    if log_dir_path is None and (log_directory_name is None or not log_directory_name.startswith(b'/')):
        msg = 'Secure open on relative path not supported'
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)
    if log_dir_path is None and (flags & os.O_DIRECTORY) == 0:
        msg = 'Opening directory but O_DIRECTORY flag missing'
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)
    if log_dir_fd is None:
        if base_dir_path is not None and base_dir_path.startswith(os.path.split(log_directory_name)[0]):
            log_dir_fd = os.open(log_directory_name, flags | os.O_NOFOLLOW | os.O_NOCTTY | os.O_DIRECTORY, dir_fd=base_dir_fd)
            log_dir_path = log_directory_name
        else:
            log_dir_fd = os.open(log_directory_name, flags | os.O_NOFOLLOW | os.O_NOCTTY | os.O_DIRECTORY)
            log_dir_path = log_directory_name
    return log_dir_fd


def close_log_directory():
    """Close the base directory at program shutdown."""
    global log_dir_fd  # skipcq: PYL-W0603
    global log_dir_path  # skipcq: PYL-W0603
    try:
        if log_dir_fd is not None:
            os.close(log_dir_fd)
            log_dir_fd = None
            log_dir_path = None
    except OSError as e:
        msg = 'Could not close the base log directory. Error: %s' % e
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)


def secure_open_file(file_name, flags):
    """
    Secure opening of a file with given flags. This call will refuse to open files where any path component is a symlink.
    As operating system does not provide any means to do that, open the file_name directory by directory. It also adds O_NOCTTY to the
    flags as controlling TTY logics as this is just an additional risk and does not make sense for opening of log files.
    @param file_name is the file name as byte string
    """
    if isinstance(file_name, str):
        file_name = file_name.encode()
    if not file_name.startswith(b'/'):
        msg = 'Secure open on relative path not supported'
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)
    if (file_name.endswith(b'/')) and ((flags & os.O_DIRECTORY) == 0):
        msg = 'Opening directory but O_DIRECTORY flag missing'
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)

    global base_dir_path  # skipcq: PYL-W0603, PYL-W0602
    global base_dir_fd  # skipcq: PYL-W0603, PYL-W0602
    if base_dir_path is not None:
        if file_name.startswith(base_dir_path):
            base_name = file_name.replace(base_dir_path, b'').lstrip(b'/')
        else:
            base_name = file_name
        return os.open(base_name, flags | os.O_NOFOLLOW | os.O_NOCTTY, dir_fd=base_dir_fd)
    dir_name = os.path.dirname(file_name)
    base_name = os.path.basename(file_name)
    dir_fd = os.open(dir_name, flags | os.O_NOFOLLOW | os.O_NOCTTY | os.O_DIRECTORY)
    ret_fd = os.open(base_name, flags | os.O_NOFOLLOW | os.O_NOCTTY, dir_fd=dir_fd)
    os.close(dir_fd)
    return ret_fd


def send_annotated_file_descriptor(send_socket, send_fd, type_info, annotation_data):
    """
    Send file descriptor and associated annotation data via SCM_RIGHTS.
    @param type_info has to be a null-byte free string to inform the receiver how to handle the file descriptor and how to interpret
    the annotationData.
    @param annotation_data this optional byte array  may convey additional information about the file descriptor.
    """
    # Construct the message data first
    if isinstance(type_info, str):
        type_info = type_info.encode()
    if isinstance(annotation_data, str):
        annotation_data = annotation_data.encode()
    if type_info.find(b'\x00') >= 0:
        msg = 'Null bytes not supported in typeInfo'
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)
    message_data = b'%s\x00%s' % (type_info, annotation_data)
    send_socket.sendmsg([message_data], [(socket.SOL_SOCKET, socket.SCM_RIGHTS, struct.pack('i', send_fd))])


def send_logstream_descriptor(send_socket, send_fd, send_file_name):
    """Send a file descriptor to be used as standard log data stream source for the analysis pipeline."""
    send_annotated_file_descriptor(send_socket, send_fd, b'logstream', send_file_name)


def receive_annoted_file_descriptor(receive_socket):
    """
    Receive a single file descriptor and attached annotation information via SCM_RIGHTS via the given socket.
    The method may raise an Exception when invoked on non-blocking sockets and no messages available.
    @return a tuple containing the received file descriptor, type information (see sendAnnotatedFileDescriptor) and the annotation
    information.
    """
    message_data, anc_data, _flags, _remote_address = receive_socket.recvmsg(1 << 16, socket.CMSG_LEN(struct.calcsize('i')))
    if len(anc_data) != 1:
        msg = 'Received %d sets of ancillary data instead of 1' % len(anc_data)
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)
    cmsg_level, cmsg_type, cmsg_data = anc_data[0]
    if (cmsg_level != socket.SOL_SOCKET) or (cmsg_type != socket.SCM_RIGHTS):
        msg = 'Received invalid message from remote side'
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)
    # Do not accept multiple or unaligned FDs.
    if len(cmsg_data) != 4:
        msg = 'Unsupported control message length %d' % len(cmsg_data)
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)
    received_fd = struct.unpack('i', cmsg_data)[0]

    split_pos = message_data.find(b'\x00')
    if split_pos < 0:
        msg = 'No null byte in received message'
        logging.getLogger(DEBUG_LOG_NAME).error(msg)
        raise Exception(msg)
    type_info = message_data[:split_pos]
    annotation_data = message_data[split_pos + 1:]
    if received_fd <= 2:
        msg = 'received "reserved" fd %d' % received_fd
        logging.getLogger(DEBUG_LOG_NAME).warning(msg)
        print('WARNING: ' + msg, file=sys.stderr)
    if isinstance(type_info, str):
        type_info = type_info.encode()
    if isinstance(annotation_data, str):
        annotation_data = annotation_data.encode()
    return received_fd, type_info, annotation_data



"""This module contains functions and classes to create the parsing model."""

from aminer.parsing.AnyByteDataModelElement import AnyByteDataModelElement
from aminer.parsing.DecimalIntegerValueModelElement import DecimalIntegerValueModelElement
from aminer.parsing.DecimalFloatValueModelElement import DecimalFloatValueModelElement
from aminer.parsing.DelimitedDataModelElement import DelimitedDataModelElement
from aminer.parsing.ElementValueBranchModelElement import ElementValueBranchModelElement
from aminer.parsing.FirstMatchModelElement import FirstMatchModelElement
from aminer.parsing.FixedDataModelElement import FixedDataModelElement
from aminer.parsing.FixedWordlistDataModelElement import FixedWordlistDataModelElement
from aminer.parsing.HexStringModelElement import HexStringModelElement
from aminer.parsing.IpAddressDataModelElement import IpAddressDataModelElement
from aminer.parsing.MatchElement import MatchElement
from aminer.parsing.OptionalMatchModelElement import OptionalMatchModelElement
from aminer.parsing.RepeatedElementDataModelElement import RepeatedElementDataModelElement
from aminer.parsing.SequenceModelElement import SequenceModelElement
from aminer.parsing.VariableByteDataModelElement import VariableByteDataModelElement
from aminer.parsing.WhiteSpaceLimitedDataModelElement import WhiteSpaceLimitedDataModelElement
from aminer.parsing.ModelElementInterface import ModelElementInterface


def get_model():
    """Return a model to parse a audispd message logged via syslog after any standard logging preamble, e.g. from syslog."""

    class ExecArgumentDataModelElement(ModelElementInterface):
        """This is a helper class for parsing the (encoded) exec argument strings found within audit logs."""

        def get_match_element(self, target_path: str, match_context):
            """
            Find the maximum number of bytes belonging to an exec argument.
            @return a match when at least two bytes were found including the delimiters.
            """
            data = match_context.match_data
            match_len = 0
            match_value = b""
            if data[0] == ord(b'"'):
                match_len = data.find(b'"', 1)
                if match_len == -1:
                    return None
                match_value = data[1:match_len]
                match_len += 1
            elif data.startswith(b"(null)"):
                match_len = 6
                match_value = None
            else:
                # Must be upper case hex encoded:
                next_value = -1
                for d_byte in data:
                    if 0x30 <= d_byte <= 0x39:
                        d_byte -= 0x30
                    elif 0x41 <= d_byte <= 0x46:
                        d_byte -= 0x37
                    else:
                        break
                    if next_value == -1:
                        next_value = (d_byte << 4)
                    else:
                        match_value += bytearray(((next_value | d_byte),))
                        next_value = -1
                    match_len += 1
                if next_value != -1:
                    return None

            match_data = data[:match_len]
            match_context.update(match_data)
            return MatchElement(f"{target_path}/{self.element_id}", match_data, match_value, None)

    pam_status_word_list = FixedWordlistDataModelElement("status", [b"failed", b"success"])

    pid = b" pid="
    uid = b" uid="
    auid = b" auid="
    gid = b" gid="
    ses = b" ses="
    exe = b' exe="'
    hostname = b'" hostname='
    hostname1 = b'" (hostname='
    addr = b" addr="
    addr1 = b", addr="
    terminal = b" terminal="
    terminal1 = b", terminal="
    res = b" res="
    exe1 = b'" exe="'
    subj = b" subj="
    comm = b" comm="
    reason = b" reason="
    dev = b" dev="
    sig = b" sig="
    alphabet = b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789._-"
    perms_alphabet = b"abcdefghijklmnopqrstuvwxyz,"

    type_branches = {
        "ADD_GROUP": SequenceModelElement("addgroup", [
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", ses),
            DecimalIntegerValueModelElement("ses"),
            FixedDataModelElement("s4", subj),
            DelimitedDataModelElement("subj", b" "),
            FixedDataModelElement("s5", b" msg='op=adding group acct=\""),
            DelimitedDataModelElement("acct", b'"'),
            FixedDataModelElement("s6", b'"'),
            FixedDataModelElement("s7", exe),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s8", hostname),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s9", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s10", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s11", res),
            pam_status_word_list,
            FixedDataModelElement("s12", b"'"),
        ]),
        "ADD_USER": SequenceModelElement("adduser", [
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", ses),
            DecimalIntegerValueModelElement("ses"),
            FixedDataModelElement("s4", subj),
            DelimitedDataModelElement("subj", b" "),
            FixedWordlistDataModelElement("s5", [b" msg='op=adding user id=", b" msg='op=adding home directory id="]),
            DecimalIntegerValueModelElement("newuserid"),
            FixedDataModelElement("s6", exe),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s7", hostname),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s8", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s9", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s10", res),
            pam_status_word_list,
            FixedDataModelElement("s11", b"'"),
        ]),
        "ANOM_ABEND": SequenceModelElement("anom_abend", [
            FixedDataModelElement("s0", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", gid),
            DecimalIntegerValueModelElement("gid"),
            FixedDataModelElement("s3", ses),
            DecimalIntegerValueModelElement("ses"),
            FixedDataModelElement("s4", subj),
            DelimitedDataModelElement("subj", b" "),
            FixedDataModelElement("s5", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s6", comm),
            ExecArgumentDataModelElement("command"),
            FixedDataModelElement("s7", reason),
            ExecArgumentDataModelElement("reason"),
            FixedDataModelElement("s8", sig),
            DecimalIntegerValueModelElement("sig")
        ]),
        "ANOM_ACCESS_FS": AnyByteDataModelElement("anom_access_fs"),
        "ANOM_ADD_ACCT": AnyByteDataModelElement("anom_add_acct"),
        "ANOM_AMTU_FAIL": AnyByteDataModelElement("anom_amtu_fail"),
        "ANOM_CRYPTO_FAIL": AnyByteDataModelElement("anom_crypto_fail"),
        "ANOM_DEL_ACCT": AnyByteDataModelElement("anom_del_acct"),
        "ANOM_EXEC": SequenceModelElement("anom_exec", [
            FixedDataModelElement("space", b" "),
            VariableByteDataModelElement("user", alphabet),
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", ses),
            DecimalIntegerValueModelElement("ses"),
            FixedDataModelElement("s4", b" msg='op="),
            DelimitedDataModelElement("msg", b" "),
            FixedDataModelElement("s5", b' acct="'),
            DelimitedDataModelElement("acct", b'"'),
            FixedDataModelElement("s6", exe1),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s7", hostname1),
            DelimitedDataModelElement("hostname", b","),
            FixedDataModelElement("s8", addr1),
            DelimitedDataModelElement("addr", b","),
            FixedDataModelElement("s9", terminal1),
            DelimitedDataModelElement("terminal", b" "),
            FixedDataModelElement("s10", res),
            pam_status_word_list,
            FixedDataModelElement("s11", b")'")
        ]),
        "ANOM_LOGIN_ACCT": AnyByteDataModelElement("anom_login_acct"),
        "ANOM_LOGIN_FAILURES": AnyByteDataModelElement("anom_login_failures"),
        "ANOM_LOGIN_LOCATION": AnyByteDataModelElement("anom_login_location"),
        "ANOM_LOGIN_SESSIONS": AnyByteDataModelElement("anom_login_sessions"),
        "ANOM_LOGIN_TIME": AnyByteDataModelElement("anom_login_time"),
        "ANOM_MAX_DAC": AnyByteDataModelElement("anom_max_dac"),
        "ANOM_MAX_MAC": AnyByteDataModelElement("anom_max_mac"),
        "ANOM_MK_EXEC": AnyByteDataModelElement("anom_mk_exec"),
        "ANOM_MOD_ACCT": AnyByteDataModelElement("anom_mod_acct"),
        "ANOM_PROMISCUOUS": SequenceModelElement("anom_promiscuous", [
            FixedDataModelElement("s0", b" dev="),
            VariableByteDataModelElement("dev", alphabet),
            FixedDataModelElement("s1", b" prom="),
            DecimalIntegerValueModelElement("prom"),
            FixedDataModelElement("s2", b" old_prom="),
            DecimalIntegerValueModelElement("old_prom"),
            FixedDataModelElement("s3", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s4", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s5", gid),
            DecimalIntegerValueModelElement("gid"),
            FixedDataModelElement("s6", ses),
            DecimalIntegerValueModelElement("ses"),
        ]),
        "ANOM_RBAC_FAIL": AnyByteDataModelElement("anom_rbac_fail"),
        "ANOM_RBAC_INTEGRITY_FAIL": AnyByteDataModelElement("anom_rbac_integrity_fail"),
        "ANOM_ROOT_TRANS": AnyByteDataModelElement("anom_root_trans"),
        "AVC": AnyByteDataModelElement("avc"),
        "AVC_PATH": AnyByteDataModelElement("avc_path"),
        "BPRM_FCAPS": SequenceModelElement("bprmfcaps", [
            FixedDataModelElement("s0", b" fver="),
            DecimalIntegerValueModelElement("fver"),
            FixedDataModelElement("s1", b" fp="),
            HexStringModelElement("fp"),
            FixedDataModelElement("s2", b" fi="),
            HexStringModelElement("fi"),
            FixedDataModelElement("s3", b" fe="),
            HexStringModelElement("fe"),
            FixedDataModelElement("s4", b" old_pp="),
            DelimitedDataModelElement("pp-old", b" "),
            FixedDataModelElement("s5", b" old_pi="),
            DelimitedDataModelElement("pi-old", b' '),
            FixedDataModelElement("s6", b" old_pe="),
            DelimitedDataModelElement("pe-old", b" "),
            FixedDataModelElement("s7", b" new_pp="),
            DelimitedDataModelElement("pp-new", b" "),
            FixedDataModelElement("s8", b" new_pi="),
            DelimitedDataModelElement("pi-new", b" "),
            FixedDataModelElement("s9", b" new_pe="),
            AnyByteDataModelElement("pe-new")
        ]),
        "CAPSET": AnyByteDataModelElement("capset"),
        "CHGRP_ID": AnyByteDataModelElement("chgrp_id"),
        "CHUSER_ID": AnyByteDataModelElement("chuser_id"),
        "CONFIG_CHANGE": SequenceModelElement("conf-change", [
            FixedDataModelElement("s0", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s1", ses),
            DecimalIntegerValueModelElement("ses"),
            FixedDataModelElement("s2", b' op="'),
            DelimitedDataModelElement("op", b'"'),
            FixedDataModelElement("s3", b'" path="'),
            DelimitedDataModelElement("path", b'"'),
            FixedDataModelElement("s4", b'" key='),
            DelimitedDataModelElement("key", b" "),
            FixedDataModelElement("s5", b' list='),
            DecimalIntegerValueModelElement("list"),
            FixedDataModelElement("s6", res),
            DecimalIntegerValueModelElement("result")
        ]),
        "CRED_ACQ": SequenceModelElement("credacq", [
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", ses),
            DecimalIntegerValueModelElement("ses"),
            FixedDataModelElement("s4", b' msg=\'op=PAM:setcred acct="'),
            DelimitedDataModelElement("username", b'"'),
            FixedDataModelElement("s5", exe1),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s6", hostname),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s7", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s8", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s9", res),
            pam_status_word_list,
            FixedDataModelElement("s10", b"'"),
        ]),
        "CRED_DISP": SequenceModelElement("creddisp", [
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", ses),
            DecimalIntegerValueModelElement("ses"),
            FixedDataModelElement("s4", b' msg=\'op=PAM:setcred acct="'),
            DelimitedDataModelElement("username", b'"'),
            FixedDataModelElement("s5", exe1),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s6", hostname),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s7", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s8", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s9", res),
            pam_status_word_list,
            FixedDataModelElement("s10", b"'"),
        ]),
        "CRED_REFR": SequenceModelElement("creddisp", [
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", ses),
            DecimalIntegerValueModelElement("ses"),
            FixedDataModelElement("s4", b' msg=\'op=PAM:setcred acct="root" exe="/usr/sbin/sshd" hostname='),
            IpAddressDataModelElement("clientname"),
            FixedDataModelElement("s5", addr),
            IpAddressDataModelElement("clientip"),
            FixedDataModelElement("s6", terminal),
            DelimitedDataModelElement("terminal", b" "),
            FixedDataModelElement("s7", res),
            pam_status_word_list,
            FixedDataModelElement("s8", b"'"),
        ]),
        "CRYPTO_FAILURE_USER": AnyByteDataModelElement("crypto_failure_user"),
        "CRYPTO_KEY_USER": AnyByteDataModelElement("crypto_key_user"),
        "CRYPTO_LOGIN": AnyByteDataModelElement("crypto_login"),
        "CRYPTO_LOGOUT": AnyByteDataModelElement("crypto_logout"),
        "CRYPTO_PARAM_CHANGE_USER": AnyByteDataModelElement("crypto_param_change_user"),
        "CRYPTO_REPLAY_USER": AnyByteDataModelElement("crypto_replay_user"),
        "CRYPTO_SESSION": SequenceModelElement("crypto_session", [
            FixedDataModelElement("space", b" "),
            VariableByteDataModelElement("user", alphabet),
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s4", b" msg='op="),
            DelimitedDataModelElement("msg", b" "),
            FixedDataModelElement("s5", b' direction='),
            DelimitedDataModelElement("direction", b' '),
            FixedDataModelElement("s6", b' cipher='),
            DelimitedDataModelElement("cipher", b' '),
            FixedDataModelElement("s7", b' ksize='),
            DecimalIntegerValueModelElement("ksize"),
            FixedDataModelElement("s8", b' rport='),
            DecimalIntegerValueModelElement("rport"),
            FixedDataModelElement("s9", b' laddr='),
            IpAddressDataModelElement("laddr"),
            FixedDataModelElement("s10", b' lport='),
            DecimalIntegerValueModelElement("lport"),
            FixedDataModelElement("s11", b' id='),
            DecimalIntegerValueModelElement("id"),
            FixedDataModelElement("s12", exe),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s13", hostname1),
            DelimitedDataModelElement("hostname", b","),
            FixedDataModelElement("s14", addr1),
            DelimitedDataModelElement("addr", b","),
            FixedDataModelElement("s15", terminal1),
            DelimitedDataModelElement("terminal", b" "),
            FixedDataModelElement("s16", res),
            pam_status_word_list,
            FixedDataModelElement("s17", b")'")
        ]),
        "CRYPTO_TEST_USER": AnyByteDataModelElement("crypto_test_user"),
        "CWD": SequenceModelElement("cwd", [
            FixedDataModelElement("s0", b"  cwd="),
            ExecArgumentDataModelElement("cwd")]),
        "DAC_CHECK": AnyByteDataModelElement("dac_check"),
        "DAEMON_ABORT": SequenceModelElement("daemon_abort", [
            FixedDataModelElement("s0", b" auditd error halt,"),
            FixedDataModelElement("s1", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s2", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s3", res),
            pam_status_word_list
        ]),
        "DAEMON_ACCEPT": AnyByteDataModelElement("daemon_accept"),
        "DAEMON_CLOSE": AnyByteDataModelElement("daemon_close"),
        "DAEMON_CONFIG": SequenceModelElement("daemon_config", [
            FixedDataModelElement("s0", b" auditd error getting hup info - no change, sending"),
            FixedDataModelElement("s1", auid),
            DelimitedDataModelElement("auid", b" "),
            FixedDataModelElement("s2", pid),
            DelimitedDataModelElement("pid", b" "),
            FixedDataModelElement("s3", subj),
            DelimitedDataModelElement("subj", b" "),
            FixedDataModelElement("s4", res),
            pam_status_word_list
        ]),
        "DAEMON_END": SequenceModelElement("daemon_end", [
            FixedDataModelElement("s0", b" auditd normal halt, sending"),
            FixedDataModelElement("s1", auid),
            DelimitedDataModelElement("auid", b" "),
            FixedDataModelElement("s2", pid),
            DelimitedDataModelElement("pid", b" "),
            FixedDataModelElement("s3", subj),
            OptionalMatchModelElement("optional_subj", DelimitedDataModelElement("subj", b" ")),
            FixedDataModelElement("s4", res),
            pam_status_word_list
        ]),
        "DAEMON_RESUME": SequenceModelElement("daemon_resume", [
            FixedDataModelElement("s0", b" auditd resuming logging, sending"),
            FixedDataModelElement("s1", auid),
            DelimitedDataModelElement("auid", b" "),
            FixedDataModelElement("s2", pid),
            DelimitedDataModelElement("pid", b" "),
            FixedDataModelElement("s3", subj),
            DelimitedDataModelElement("subj", b" "),
            FixedDataModelElement("s4", res),
            pam_status_word_list
        ]),
        "DAEMON_ROTATE": AnyByteDataModelElement("daemon_rotate"),
        "DAEMON_START": SequenceModelElement("daemon_start", [
            FixedDataModelElement("s0", b" auditd start, ver="),
            DecimalFloatValueModelElement("ver"),
            FixedDataModelElement("s1", b" format="),
            DelimitedDataModelElement("format", b" "),
            FixedDataModelElement("s2", b" kernel="),
            DelimitedDataModelElement("kernel", b" "),
            FixedDataModelElement("s3", auid),
            DelimitedDataModelElement("auid", b" "),
            FixedDataModelElement("s4", pid),
            DelimitedDataModelElement("pid", b" "),
            FixedDataModelElement("s5", res),
            pam_status_word_list
        ]),
        "DEL_GROUP": AnyByteDataModelElement("del_group"),
        "DEL_USER": AnyByteDataModelElement("del_user"),
        "EOE": AnyByteDataModelElement("eoe"),
        "EXECVE": SequenceModelElement("execve", [
            FixedDataModelElement("s0", b" argc="),
            DecimalIntegerValueModelElement("argc"),
            # We need a type branch here also, but there is no additional data in EOE records after Ubuntu Trusty any more.
            RepeatedElementDataModelElement("arg", SequenceModelElement("execarg", [
                FixedDataModelElement("s0", b" a"),
                DecimalIntegerValueModelElement("argn"),
                FixedDataModelElement("s1", b"="),
                ExecArgumentDataModelElement("argval")
            ]))
        ]),
        "FD_PAIR": SequenceModelElement("fdpair", [
            FixedDataModelElement("s0", b" fd0="),
            DecimalIntegerValueModelElement("fd0"),
            FixedDataModelElement("s1", b" fd1="),
            DecimalIntegerValueModelElement("fd1")
        ]),
        "FS_RELABEL": AnyByteDataModelElement("fs_relabel"),
        "GRP_AUTH": AnyByteDataModelElement("grp_auth"),
        "INTEGRITY_DATA": AnyByteDataModelElement("integrity_data"),
        "INTEGRITY_HASH": AnyByteDataModelElement("integrity_hash"),
        "INTEGRITY_METADATA": AnyByteDataModelElement("integrity_metadata"),
        "INTEGRITY_PCR": AnyByteDataModelElement("integrity_pcr"),
        "INTEGRITY_RULE": AnyByteDataModelElement("integrity_rule"),
        "INTEGRITY_STATUS": AnyByteDataModelElement("integrity_status"),
        "IPC": AnyByteDataModelElement("ipc"),
        "IPC_SET_PERM": AnyByteDataModelElement("ipc_set_perm"),
        "KERNEL": AnyByteDataModelElement("kernel"),
        "KERNEL_OTHER": AnyByteDataModelElement("kernel_other"),
        "LABEL_LEVEL_CHANGE": AnyByteDataModelElement("label_level_change"),
        "LABEL_OVERRIDE": AnyByteDataModelElement("label_override"),
        # This message differs on Ubuntu 32/64 bit variants.
        "LOGIN": SequenceModelElement("login", [
            FixedDataModelElement("s0", b" login"),
            FixedDataModelElement("s1", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s2", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedWordlistDataModelElement("s3", [b" old auid=", b" old-auid="]),
            DecimalIntegerValueModelElement("auid-old"),
            FixedWordlistDataModelElement("s4", [b" new auid=", auid]),
            DecimalIntegerValueModelElement("auid-new"),
            FixedWordlistDataModelElement("s5", [b" old ses=", b" old-ses="]),
            DecimalIntegerValueModelElement("ses-old"),
            FixedWordlistDataModelElement("s6", [b" new ses=", ses]),
            DecimalIntegerValueModelElement("ses-new"),
            OptionalMatchModelElement("optional_result", SequenceModelElement("result_seq", [
                FixedDataModelElement("s7", res),
                DecimalIntegerValueModelElement("result")
            ]))
        ]),
        "MAC_CIPSOV4_ADD": AnyByteDataModelElement("mac_cipsov4_add"),
        "MAC_CIPSOV4_DEL": AnyByteDataModelElement("mac_cipsov4_del"),
        "MAC_CONFIG_CHANGE": AnyByteDataModelElement("mac_config_change"),
        "MAC_IPSEC_EVENT": AnyByteDataModelElement("mac_ipsec_event"),
        "MAC_MAP_ADD": AnyByteDataModelElement("mac_map_add"),
        "MAC_MAP_DEL": AnyByteDataModelElement("mac_map_del"),
        "MAC_POLICY_LOAD": AnyByteDataModelElement("mac_policy_load"),
        "MAC_STATUS": SequenceModelElement("mac_status", [
            FixedDataModelElement("s0", b" enforcing="),
            DecimalIntegerValueModelElement("enforcing"),
            FixedDataModelElement("s1", b" old_enforcing="),
            DecimalIntegerValueModelElement("old_enforcing"),
            FixedDataModelElement("s2", auid),
            DelimitedDataModelElement("auid", b" "),
            FixedDataModelElement("s3", ses),
            DecimalIntegerValueModelElement("ses")
        ]),
        "MAC_UNLBL_ALLOW": AnyByteDataModelElement("mac_unlbl_allow"),
        "MAC_UNLBL_STCADD": AnyByteDataModelElement("mac_unlbl_stcadd"),
        "MAC_UNLBL_STCDEL": AnyByteDataModelElement("mac_unlbl_stcdel"),
        "MMAP": AnyByteDataModelElement("mmap"),
        "MQ_GETSETATTR": AnyByteDataModelElement("mq_getsetattr"),
        "MQ_NOTIFY": AnyByteDataModelElement("mq_notify"),
        "MQ_OPEN": AnyByteDataModelElement("mq_open"),
        "MQ_SENDRECV": AnyByteDataModelElement("mq_sendrecv"),
        "NETFILTER_CFG": SequenceModelElement("netfilter_cfg", [
            FixedDataModelElement("s0", b" table="),
            FixedWordlistDataModelElement("table", [b"filter", b"mangle", b"nat"]),
            FixedDataModelElement("s1", b" family="),
            DecimalIntegerValueModelElement("family"),
            FixedDataModelElement("s2", b" entries="),
            DecimalIntegerValueModelElement("entries")
        ]),
        "NETFILTER_PKT": SequenceModelElement("netfilter_pkt", [
            FixedDataModelElement("s0", b" mark=0x"),
            HexStringModelElement("mark"),
            FixedDataModelElement("s1", b" saddr="),
            FirstMatchModelElement("saddr", [
                IpAddressDataModelElement("ipv4"),
                IpAddressDataModelElement("ipv6", ipv6=True),
            ]),
            FixedDataModelElement("s2", b" daddr="),
            FirstMatchModelElement("daddr", [
                IpAddressDataModelElement("ipv4"),
                IpAddressDataModelElement("ipv6", ipv6=True),
            ]),
            FixedDataModelElement("s3", b" proto="),
            DecimalIntegerValueModelElement("proto")
        ]),
        "OBJ_PID": SequenceModelElement("objpid", [
            FixedDataModelElement("s0", b" opid="),
            DecimalIntegerValueModelElement("opid"),
            FixedDataModelElement("s1", b" oauid="),
            DecimalIntegerValueModelElement("oauid", value_sign_type=DecimalIntegerValueModelElement.SIGN_TYPE_OPTIONAL),
            FixedDataModelElement("s2", b" ouid="),
            DecimalIntegerValueModelElement("ouid"),
            FixedDataModelElement("s3", b" oses="),
            DecimalIntegerValueModelElement("oses", value_sign_type=DecimalIntegerValueModelElement.SIGN_TYPE_OPTIONAL),
            FixedDataModelElement("s4", b" obj="),
            DelimitedDataModelElement("obj", b" "),
            FixedDataModelElement("s4", b" ocomm="),
            ExecArgumentDataModelElement("ocomm")
        ]),
        "PATH": SequenceModelElement("path", [
            FixedDataModelElement("s0", b" item="),
            DecimalIntegerValueModelElement("item"),
            FixedDataModelElement("s1", b" name="),
            ExecArgumentDataModelElement("name"),
            FirstMatchModelElement("fsinfo", [
                SequenceModelElement("inodeinfo", [
                    FixedDataModelElement("s0", b" inode="),
                    DecimalIntegerValueModelElement("inode"),
                    FixedDataModelElement("s1", dev),
                    # A special major/minor device element could be better here.
                    VariableByteDataModelElement("dev", b"0123456789abcdef:"),
                    FixedDataModelElement("s2", b" mode="),
                    # is octal
                    DecimalIntegerValueModelElement("mode", value_pad_type=DecimalIntegerValueModelElement.PAD_TYPE_ZERO),
                    FixedDataModelElement("s3", b" ouid="),
                    DecimalIntegerValueModelElement("ouid"),
                    FixedDataModelElement("s4", b" ogid="),
                    DecimalIntegerValueModelElement("ogid"),
                    FixedDataModelElement("s5", b" rdev="),
                    # A special major/minor device element could be better here (see above).
                    VariableByteDataModelElement("rdev", b"0123456789abcdef:"),
                    FixedDataModelElement("s6", b" nametype=")
                ]),
                FixedDataModelElement("noinfo", b" nametype=")]),
            FixedWordlistDataModelElement("nametype", [b"CREATE", b"DELETE", b"NORMAL", b"PARENT", b"UNKNOWN"])
        ]),
        "PROCTITLE": SequenceModelElement("proctitle", [
            FixedDataModelElement("s0", b" proctitle="),
            ExecArgumentDataModelElement("proctitle")]),
        "RESP_ACCT_LOCK": AnyByteDataModelElement("resp_acct_lock"),
        "RESP_ACCT_LOCK_TIMED": AnyByteDataModelElement("resp_acct_lock_timed"),
        "RESP_ACCT_REMOTE": AnyByteDataModelElement("resp_acct_remote"),
        "RESP_ACCT_UNLOCK_TIMED": AnyByteDataModelElement("resp_acct_unlock_timed"),
        "RESP_ALERT": AnyByteDataModelElement("resp_alert"),
        "RESP_ANOMALY": AnyByteDataModelElement("resp_anomaly"),
        "RESP_EXEC": AnyByteDataModelElement("resp_exec"),
        "RESP_HALT": AnyByteDataModelElement("resp_halt"),
        "RESP_KILL_PROC": AnyByteDataModelElement("resp_kill_proc"),
        "RESP_SEBOOL": AnyByteDataModelElement("resp_sebool"),
        "RESP_SINGLE": AnyByteDataModelElement("resp_single"),
        "RESP_TERM_ACCESS": AnyByteDataModelElement("resp_term_access"),
        "RESP_TERM_LOCK": AnyByteDataModelElement("resp_term_lock"),
        "ROLE_ASSIGN": AnyByteDataModelElement("role_assign"),
        "ROLE_MODIFY": AnyByteDataModelElement("role_modify"),
        "ROLE_REMOVE": AnyByteDataModelElement("role_remove"),
        "SELINUX_ERR": SequenceModelElement("service_err", [
            FixedDataModelElement("s0", b" op="),
            DelimitedDataModelElement("op", b" "),
            FixedDataModelElement("s1", reason),
            DelimitedDataModelElement("reason", b" "),
            FixedDataModelElement("s2", b" scontext="),
            DelimitedDataModelElement("scontext", b" "),
            FixedDataModelElement("s3", b" tcontext="),
            DelimitedDataModelElement("tcontext", b" "),
            FixedDataModelElement("s4", b" tclass="),
            DelimitedDataModelElement("tclass", b" "),
            FixedDataModelElement("s5", b" perms="),
            VariableByteDataModelElement("perms", perms_alphabet)
        ]),
        "SERVICE_START": SequenceModelElement("service", [
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", ses),
            DecimalIntegerValueModelElement("ses"),
            FixedDataModelElement("s4", b" msg='"),
            OptionalMatchModelElement("optional_msg", DelimitedDataModelElement("msg", b" ")),
            FixedDataModelElement("s5", b' comm="'),
            DelimitedDataModelElement("comm", b'"'),
            FixedDataModelElement("s5", b'" exe="'),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s6", hostname),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s7", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s8", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s9", res),
            pam_status_word_list,
            FixedDataModelElement("s10", b"'")
        ]),
        "SOCKADDR": SequenceModelElement("sockaddr", [
            FixedDataModelElement("s0", b" saddr="),
            HexStringModelElement("sockaddr", upper_case=True)
        ]),
        "SOCKETCALL": SequenceModelElement("socketcall", [
            FixedDataModelElement("s0", b" nargs="),
            DecimalIntegerValueModelElement("nargs"),
            RepeatedElementDataModelElement("args", SequenceModelElement("arg", [
                FixedDataModelElement("s1", b" a"),
                DecimalIntegerValueModelElement("arg_num"),
                FixedDataModelElement("s2", b"="),
                DecimalIntegerValueModelElement("arg"),
            ]))
        ]),
        "SYSCALL": SequenceModelElement("syscall", [
            FixedDataModelElement("s0", b" arch="),
            HexStringModelElement("arch"),
            FixedDataModelElement("s1", b" syscall="),
            DecimalIntegerValueModelElement("syscall"),
            OptionalMatchModelElement(
                "personality", SequenceModelElement("pseq", [
                    FixedDataModelElement("s0", b" per="),
                    DecimalIntegerValueModelElement("personality")
                ])),
            OptionalMatchModelElement("result", SequenceModelElement("rseq", [
                FixedDataModelElement("s2", b" success="),
                FixedWordlistDataModelElement("succes", [b"no", b"yes"]),
                FixedDataModelElement("s3", b" exit="),
                DecimalIntegerValueModelElement("exit", value_sign_type=DecimalIntegerValueModelElement.SIGN_TYPE_OPTIONAL)
            ])),
            FixedDataModelElement("s4", b" a0="),
            HexStringModelElement("arg0"),
            FixedDataModelElement("s5", b" a1="),
            HexStringModelElement("arg1"),
            FixedDataModelElement("s6", b" a2="),
            HexStringModelElement("arg2"),
            FixedDataModelElement("s7", b" a3="),
            HexStringModelElement("arg3"),
            FixedDataModelElement("s8", b" items="),
            DecimalIntegerValueModelElement("items"),
            FixedDataModelElement("s9", b" ppid="),
            DecimalIntegerValueModelElement("ppid"),
            FixedDataModelElement("s10", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s11", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s12", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s13", gid),
            DecimalIntegerValueModelElement("gid"),
            FixedDataModelElement("s14", b" euid="),
            DecimalIntegerValueModelElement("euid"),
            FixedDataModelElement("s15", b" suid="),
            DecimalIntegerValueModelElement("suid"),
            FixedDataModelElement("s16", b" fsuid="),
            DecimalIntegerValueModelElement("fsuid"),
            FixedDataModelElement("s17", b" egid="),
            DecimalIntegerValueModelElement("egid"),
            FixedDataModelElement("s18", b" sgid="),
            DecimalIntegerValueModelElement("sgid"),
            FixedDataModelElement("s19", b" fsgid="),
            DecimalIntegerValueModelElement("fsgid"),
            FixedDataModelElement("s20", b" tty="),
            DelimitedDataModelElement("tty", b" "),
            FixedDataModelElement("s21", ses),
            DecimalIntegerValueModelElement("sesid"),
            FixedDataModelElement("s22", comm),
            ExecArgumentDataModelElement("command"),
            FixedDataModelElement("s23", exe),
            DelimitedDataModelElement("executable", b'"'),
            FixedDataModelElement("s24", b'" key='),
            AnyByteDataModelElement("key")
        ]),
        "SYSTEM_BOOT": AnyByteDataModelElement("system_boot"),
        "SYSTEM_RUNLEVEL": AnyByteDataModelElement("system_runlevel"),
        "SYSTEM_SHUTDOWN": AnyByteDataModelElement("system_shutdown"),
        "TRUSTED_APP": AnyByteDataModelElement("trusted_app"),
        "TTY": AnyByteDataModelElement("tty"),
        # The UNKNOWN type is used then audispd does not know the type of the event, usually because the kernel is more recent than audispd,
        # thus emiting yet unknown event types.
        # * type=1327: procitle: see https://www.redhat.com/archives/linux-audit/2014-February/msg00047.html
        "UNKNOWN[1327]": SequenceModelElement("unknown-proctitle", [
            FixedDataModelElement("s0", b" proctitle="),
            ExecArgumentDataModelElement("proctitle")
        ]),
        "USER_ACCT": SequenceModelElement("useracct", [
            FixedDataModelElement("space", b" "),
            DelimitedDataModelElement("user", b" "),
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", ses),
            DecimalIntegerValueModelElement("ses"),
            FixedDataModelElement("s4", b' msg=\'op=PAM:accounting acct="'),
            DelimitedDataModelElement("username", b'"'),
            FixedDataModelElement("s5", exe1),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s6", hostname1),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s7", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s8", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s9", res),
            pam_status_word_list,
            FixedDataModelElement("s10", b")'")
        ]),
        "USER_AUTH": SequenceModelElement("userauth", [
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", ses),
            DecimalIntegerValueModelElement("ses"),
            FixedDataModelElement("s4", b' msg=\'op=PAM:authentication acct="'),
            DelimitedDataModelElement("username", b'"'),
            FixedDataModelElement("s5", exe1),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s6", hostname),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s7", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s8", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s9", res),
            pam_status_word_list,
            FixedDataModelElement("s10", b"'")
        ]),
        "USER_AVC": AnyByteDataModelElement("user_avc"),
        "USER_CHAUTHTOK": AnyByteDataModelElement("user_chauthtok"),
        "USER_CMD": SequenceModelElement("user_cmd", [
            FixedDataModelElement("space", b" "),
            DelimitedDataModelElement("user", b" "),
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", b" msg='"),
            DelimitedDataModelElement("msg", b" "),
            FixedDataModelElement("s4", b' cmd="'),
            DelimitedDataModelElement("cmd", b'"'),
            FixedDataModelElement("s5", b"\" (terminal=pts/0"),
            FixedDataModelElement("s6", res),
            pam_status_word_list,
            FixedDataModelElement("s7", b")'"),
        ]),
        "USER_END": SequenceModelElement("userend", [
            FixedDataModelElement("space", b" "),
            DelimitedDataModelElement("user", b" "),
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s4", b' msg=\'PAM: session close acct="'),
            DelimitedDataModelElement("username", b'"'),
            FixedDataModelElement("s5", b'" :' + exe),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s6", hostname1),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s7", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s8", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s9", res),
            pam_status_word_list,
            FixedDataModelElement("s10", b")'"),
        ]),
        "USER_ERR": SequenceModelElement("usererr", [
            FixedDataModelElement("space", b" "),
            DelimitedDataModelElement("user", b" "),
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", b' msg=\'PAM: bad_ident acct=? : exe="'),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s4", hostname1),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s5", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s6", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s7", res),
            pam_status_word_list,
            FixedDataModelElement("s8", b")'")
        ]),
        "USER_LABELED_EXPORT": AnyByteDataModelElement("user_labeled_export"),
        "USER_LOGIN": SequenceModelElement("userlogin", [
            FixedDataModelElement("space", b" "),
            DelimitedDataModelElement("user", b" "),
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", b" msg='acct=\""),
            DelimitedDataModelElement("acct", b'"'),
            FixedDataModelElement("s4", b'":' + exe),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s5", hostname1),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s6", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s7", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s8", res),
            pam_status_word_list,
            FixedDataModelElement("s9", b")'")
        ]),
        "USER_LOGOUT": AnyByteDataModelElement("user_logout"),
        "USER_MAC_POLICY_LOAD": AnyByteDataModelElement("user_mac_policy_load"),
        "USER_MGMT": AnyByteDataModelElement("user_mgmt"),
        "USER_ROLE_CHANGE": SequenceModelElement("user_role_change", [
            FixedDataModelElement("space", b" "),
            DelimitedDataModelElement("user", b" "),
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", subj),
            DelimitedDataModelElement("subj", b" "),
            FixedDataModelElement("s4", b" msg='pam: "),
            DelimitedDataModelElement("msg", b" "),
            FixedDataModelElement("s5", b" selected-context="),
            DelimitedDataModelElement("selected_context", b" "),
            FixedDataModelElement("s6", exe),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s7", hostname1),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s8", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s9", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s10", res),
            pam_status_word_list,
            FixedDataModelElement("s11", b")'")
        ]),
        "USER_SELINUX_ERR": AnyByteDataModelElement("user_selinux_err"),
        "USER_START": SequenceModelElement("userstart", [
            FixedDataModelElement("space", b" "),
            DelimitedDataModelElement("user", b" "),
            FixedDataModelElement("s0", pid),
            DecimalIntegerValueModelElement("pid"),
            FixedDataModelElement("s1", uid),
            DecimalIntegerValueModelElement("uid"),
            FixedDataModelElement("s2", auid),
            DecimalIntegerValueModelElement("auid"),
            FixedDataModelElement("s3", b' msg=\'PAM: session open acct="'),
            DelimitedDataModelElement("username", b'"'),
            FixedDataModelElement("s4", b'" :' + exe),
            DelimitedDataModelElement("exec", b'"'),
            FixedDataModelElement("s5", hostname1),
            DelimitedDataModelElement("clientname", b" "),
            FixedDataModelElement("s6", addr),
            DelimitedDataModelElement("clientip", b" "),
            FixedDataModelElement("s7", terminal),
            WhiteSpaceLimitedDataModelElement("terminal"),
            FixedDataModelElement("s8", res),
            pam_status_word_list,
            FixedDataModelElement("s9", b")'"),
        ]),
        "USER_TTY": AnyByteDataModelElement("user_tty"),
        "USER_UNLABELED_EXPORT": AnyByteDataModelElement("user_unlabeled_export"),
        "USYS_CONFIG": AnyByteDataModelElement("usys_config"),
        "VIRT_CONTROL": AnyByteDataModelElement("virt_control"),
        "VIRT_MACHINE_ID": AnyByteDataModelElement("virt_machine_id"),
        "VIRT_RESOURCE": AnyByteDataModelElement("virt_resource")
    }

    type_branches["SERVICE_STOP"] = type_branches["SERVICE_START"]

    model = SequenceModelElement("audispd", [
        OptionalMatchModelElement("optional", FirstMatchModelElement("type", [
            FixedDataModelElement("sname", b"audispd: "),
            FixedDataModelElement("sname_remote", b"audisp-remote: "),
        ])),
        FirstMatchModelElement("msg", [
            ElementValueBranchModelElement("record", SequenceModelElement("preamble", [
                FixedDataModelElement("s0", b"type="),
                WhiteSpaceLimitedDataModelElement("type"),
                FixedDataModelElement("s1", b" msg=audit("),
                DecimalIntegerValueModelElement("time"),
                FixedDataModelElement("s0", b"."),
                DecimalIntegerValueModelElement("ms", value_pad_type=DecimalIntegerValueModelElement.PAD_TYPE_ZERO),
                FixedDataModelElement("s1", b":"),
                DecimalIntegerValueModelElement("seq"),
                FixedDataModelElement("s2", b"):")
            ]), "type", type_branches, default_branch=None),
            FixedDataModelElement("queue-full", b"queue is full - dropping event")
        ])
    ])
    return model
